#define HAVE_UNISTD_H
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <ftw.h>
                     
#include "lib/libdigest/digest.h"
#include "lib/safe-rw.h"
#include "lib/shell.h"
#include "lib/error.h"

#ifndef FTW_SL
#define FTW_SL 0
#ifndef FTW_PHYS
#define FTW_PHYS 0
#endif
struct FTW {
}         ;
#define nftw(a,b,c,d) ftw(a,b,c)
#define lstat stat
#define getopt_long(a,b,c,d,e) getopt(a,b,c)
#define use_backslash
#endif

#ifndef S_IFLNK
#define S_IFLNK -1
#endif

struct opts_t {
	int base64;
	int debug;
	int flinks;
	int sortable;
	char * output;
};

struct opts_t opts={0};

struct option longoptions[]=
{
	{"hash",1,NULL,0},
	{"base64",0,&opts.base64,1},
	{"debug",0,&opts.debug,1},
	{"out",1,NULL,0},
	{"follow-links",1,&opts.flinks,1},
	{"help",0,NULL,0},
	{"sortable",0,&opts.sortable,1},
	{NULL,0,NULL,0}
};

static char help[]=
"dglist: create a list of files and their digest\n"
"\n"
"Options:\n"
"  --hash       <hash>     -H  include hash\n"
"  --out        <file>     -o  send output to <file>\n"
"  --follow-symlinks       -f  follow symlinks, possibly ending in a loop\n"
"  --base64                -b  encode hashes in base64\n"
"  --sortable              -s  print the hashes first\n"
"  --debug                 -d  generate debugging output\n"
"  --help                  -h  this text\n";

char options[]="H:o:bdfhsW;";

char hex[16]="0123456789abcdef";

int digest_idx(char * name)
{
	int i;
	for(i=0;i<digest_n;i++) {
		if(!strcmp(name,digest[i].name)) return(i);
	}
	return(-1);
}

void digest_list()
{
	int i;
	fprintf(stderr,"valid digests are: ");
	for(i=0;i<digest_n;i++) {
		fprintf(stderr," %s",digest[i].name);
	}
	fprintf(stderr,"\n");
}

void print_hb(FILE * F, int base64, unsigned char * buf, size_t len)
{
	int j,l,m;
	char base[64]="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	              "abcdefghijklmnopqrstuvwxyz0123456789+/";
	
	if(base64) {
		l=(len/3)*3;
		m=len-l;
		for(j=0;j<l;j+=3){
			fprintf(F,"%c%c%c%c",base[(buf[j]>>2) & 0x3f],
			                  base[((buf[j  ]<<4)|(buf[j+1]>>4)) & 0x3f],
			                  base[((buf[j+1]<<2)|(buf[j+2]>>6)) & 0x3f],
			                  base[  buf[j+2] & 0x3f]);
		}
		switch(m) {
		case 2:
			fprintf(F,"%c%c%c=",base[(buf[j]>>2) & 0x3f],
			                  base[((buf[j  ]<<4)|(buf[j+1]>>4)) & 0x3f],
			                  base[(buf[j+1]<<2) & 0x3f]);
			break;
		case 1:
			fprintf(F,"%c%c==",base[(buf[j]>>2) & 0x3f],
			                  base[(buf[j  ]<<4) & 0x3f]);
		}
	} else {
		for(j=0;j<len;++j){
			fprintf(F,"%c%c",hex[(buf[j] >>4)], hex[(buf[j] & 0xf)]);
}	}	}

void print_fn(FILE * F,unsigned char const * file)
{
	int c;
	while((c=*file++)){
#ifdef use_backslash
		if(c=='/')c='\\';
#endif
		if((c>'!')&&(c<='~')) fputc(c,F);
		else fprintf(F,"!%c%c",hex[c>>4], hex[c&0xf]);
}	}

int do_file(FILE * F,int digests[digest_n], char const * file)
{
	int f = open(file,O_RDONLY);
	char buf[16384];
	void * ctx[digest_n];
	unsigned char * res[digest_n];
	int r;
	int i;
	
	if(f<0) {
		qerror2(EEOPEN, "in do_file: ", file);
		return(-1);
	}

	for(i=0;i<digest_n;i++) {
		if(digests[i]){
			ctx[i]=alloca(digest[i].buflen);
			res[i]=alloca(digest[i].reslen);
			digest[i].binit(ctx[i]);
	}	}

	while((r=safe_read(f,buf,sizeof(buf)))>0) {
		for(i=0;i<digest_n;i++) {
			if(digests[i]){
				digest[i].bnext(buf,r,ctx[i]);
	}	}	}
	if(r<0) {
		qerror2(EEREAD, "in do_file: ", file);
		return(-1);
	}
	
	if(!opts.sortable)print_fn(F,file);
	for(i=0;i<digest_n;i++) {
		if(digests[i]){
			if(!opts.sortable)fputc(' ',F);
			digest[i].bdone(ctx[i],res[i]);

			fprintf(F,"%s:",digest[i].name);
			print_hb(F,opts.base64,res[i],digest[i].reslen);
			if(opts.sortable)fputc(' ',F);
		}
	}
	if(opts.sortable)print_fn(F,file);
	fputc('\n',F);

	if(close(f)) {
		qerror2(EECLOSE, "in do_file: ", file);
		return(-1);
	}
	return(0);
}

void do_arg(FILE * F,int digests[digest_n],char const * arg);

int do_dir(FILE * F, int digests[digest_n],char const * dir)
{
	int do_dir_2(const char *file, const  struct  stat  *sb,
		int flag, struct FTW *s)
	{
		switch(flag){
		case FTW_F: 	// Item is a normal file
		case FTW_SL:	// Item is a symbolic link
			do_arg(F,digests,file);
			break;
		case FTW_D: 	// Item is a directory
			// handled by nftw
			break;
		case FTW_DNR:	// Item is a directory which can't be read
			qerror2(EEREAD,"in do_dir_2: ",file);
			xerror();
			break;
		case FTW_NS:	// The stat failed on the item which is not a symbolic link
			qerror2(EESTAT,"in do_dir_2: ",file);
			xerror();
			break;
		}
		return 0;
	}

	if(nftw(dir,do_dir_2,5,opts.flinks?0:FTW_PHYS)<0) {
		qerror2(EEUNKNOWN, "in do_dir: nftw ",dir);
		return(-1);
	}
	return(0);
}

void do_arg(FILE * F,int digests[digest_n],char const * arg)
{
	struct stat st;
	lstat(arg,&st);
	switch(st.st_mode & S_IFMT){
	case S_IFLNK:
		if(!opts.flinks) {
			fprintf(stderr,"skipping symbolic link ");
			print_fn(stderr,arg);
			fputc('\n',stderr);
			break;
		}
	case S_IFREG:
		if(do_file(F,digests,arg)<0) xerror();
		break;
	case S_IFDIR:
		if(do_dir(F,digests,arg)<0) xerror();
		break;
	default:
		fprintf(stderr,"skipping special file ");
		print_fn(stderr,arg);
		fputc('\n',stderr);
		break;
	}
}

int main(int argc, char**argv)
{
	int option_index = 0;
	char c = getopt_long (argc, argv, options, longoptions, &option_index);
	int digests[digest_n];
	int i;
	char ** p;
	FILE * F;

	memset(digests,0,sizeof(digests));
	
	while(c != -1) {
		switch (c) {
		case 0:
			if(longoptions[option_index].flag) break;
			
			switch(option_index){
			case 0:	// hash
				i=digest_idx(optarg);
				if(i<0) {
					if(strcmp(optarg,"help"))
					 fprintf(stderr,"invalid hash \"%s\"\n",optarg);
					digest_list();
					exit(1);
				}
				digests[i]=1;
				break;
			case 3:
				opts.output=optarg;
				break;
			case 5:
				fprintf(stderr,help);
				exit(0);
			default:
				printf("lopt %s %s\n",longoptions[option_index].name,
				optarg?optarg:"");
				exit(1);
			}
			break;
		case 'b':
			opts.base64=1;
			break;
		case 'd':
			opts.debug=1;
			break;
		case 's':
			opts.sortable=1;
			break;
		case 'f':
			opts.flinks=1;
			break;
		case 'o':
			opts.output=optarg;
			break;
		case 'H':
			i=digest_idx(optarg);
			if(i<0) {
				if(strcmp(optarg,"help"))
				 fprintf(stderr,"invalid hash \"%s\"\n",optarg);
				digest_list();
				exit(1);
			}
			digests[i]=1;
			break;
		case 'h':
			fprintf(stderr,help);
			exit(1);
		case '?':
			exit(1);
		}
		
		c = getopt_long(argc, argv, options, longoptions, &option_index);
	}

	if(opts.debug){
		printf("options follow-symlinks %d base64 %d sortable %d debug %d\n",
		opts.flinks, opts.base64, opts.sortable, opts.debug);
	
		for(i=0;i<digest_n;i++)
		 if(digests[i]) printf("using hash %s\n",digest[i].name);
	}
	
	F=stdout;
	if(opts.output && strcmp(opts.output,"-")) {
		if(!(F=fopen(opts.output,"w"))){
			qerror2(EEOPEN,"in main: ",opts.output);
			xerror();
			exit(1);
		}
	}
	
	for(p=&argv[optind];*p;++p) {
		do_arg(F,digests,*p);
	}

	exit(0);
}
