#include <sys/types.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <stdio.h>
                     
#include "error.h"
#include "safe-rw.h"
#include "hash.h"

#define HASH_LEAF 0x80000000
#define HASH_IDC 0x000000ff

char h[16][8]={
	"03245678",
	"12345678",
	"28154711",
	"33245679",
	"42345678",
	"58154711",
	"68154711",
	"73245679",
	"12345678",
	"12345678",
	"08154711",
	"13245679",
	"12345678",
	"08154711",
	"08154711",
	"13245679"
};

static int hash_i_flush(struct hash_ctx * ctx)
{
	off_t w;
	int b=ctx->hac*sizeof(struct hash_ha_t);
	if(lseek(ctx->f,b,SEEK_SET)!=b) {
		char x[21+3*sizeof(h)];
		snprintf(x,sizeof(x),"while writing entry %d",ctx->hac);
		qerror(EESEEK,x);
		return(-1);
	};
	if((w=safe_write(ctx->f,&ctx->ha[1],sizeof(ctx->ha[1])))!=sizeof(ctx->ha[1])){
		char x[21+3*sizeof(h)];
		snprintf(x,sizeof(x),"while writing entry %d",ctx->hac);
		qerror(EEWRITE,x);
		return(-1);
	};
	ctx->hacd=0;
	return 0;
}

static int hash_i_get(struct hash_ctx * ctx, int h)
{
	if(ctx->hac!=h){
		int b=h*sizeof(struct hash_ha_t);
		if(ctx->hacd)if(hash_i_flush(ctx)<0)return(-1);
		if(h){
			if(lseek(ctx->f,b,SEEK_SET)!=b) {
				char x[21+3*sizeof(h)];
				snprintf(x,sizeof(x),"while reading entry %d",h);
				qerror(EESEEK,"x");
				return(-1);
			};
			if((safe_read(ctx->f,&ctx->ha[1],sizeof(ctx->ha[1])))!=sizeof(ctx->ha[1])){
				char x[21+3*sizeof(h)];
				snprintf(x,sizeof(x),"while reading entry %d",h);
				qerror(EEREAD,x);
				return(-1);
			};
			ctx->hac=h;
		};
	};
	return 0;
}

static int hash_i_0flush(struct hash_ctx * ctx){
	off_t w;
	if(lseek(ctx->f,0,SEEK_SET)!=0) {
		qerror(EESEEK,"in hash_i_0flush");
		return(-1);
	};
	if((w=safe_write(ctx->f,&ctx->ha[0],sizeof(ctx->ha[0])))!=sizeof(ctx->ha[0])){
		qerror(EEWRITE,"while flushing root entry");
		return(-1);
	};
	return 0;
}

int hash_insert(struct hash_ctx * ctx, off_t what1, off_t what2, uint8_t * h, int length, int hash)
{
	int a=hash*sizeof(struct hash_ha_t);
	int i;
	struct hash_hb_t ha2;
	off_t w;
	
	if(length) {
		if(hash){
			// !!hash is always 1 here
			if(hash_i_get(ctx,hash)<0)return(-1);
			if(ctx->ha[1].f & HASH_LEAF) {
				// We have a leafnode
				int foo;
				if(memcmp((*(struct hash_hb_t*)(&ctx->ha[1])).h,h,length)){
					// seperate the leaf
					int bar;
					memcpy(&ha2,&ctx->ha[1],sizeof(ha2));
					bzero(&ctx->ha[1],sizeof(ctx->ha[1]));
					ctx->hacd++;
					bar=ha2.f&HASH_IDC;
					for(i=0;i<bar;i++)
					if((foo=hash_insert(
					          ctx,
					          ha2.w1[i],
					          ha2.w2[i],
					          ha2.h,
					          length,
					          hash)) <0) {
						if(ctx->hacd)hash_i_flush(ctx);
						return(foo);
					}
					if(hash_i_get(ctx,hash)<0)return(-1);
	//	v-----------' now we can just add the entry
				} else {
					int foo;
					// we have to add a collision entry.
					// !!hash is always 1 here
					if(((foo=ctx->ha[1].f&HASH_IDC))<HASH_MAX_I){
						(*(struct hash_hb_t *)&ctx->ha[1]).w1[foo]=what1;
						(*(struct hash_hb_t *)&ctx->ha[1]).w2[foo]=what2;
						ctx->ha[1].f++;
						hash_i_flush(ctx);
						return foo;
					} else {
						// too many collisions
						return(-2);
					}
				}
			}
		}
		// just add the entry
		if((i=ctx->ha[!!hash].i[*h])) {
			return hash_insert(ctx,what1,what2,h+1,length-1,i);
		} else {
			ctx->ha[!!hash].i[*h]=ctx->han;
			if(hash)hash_i_flush(ctx);
			else hash_i_0flush(ctx);
			bzero(&ha2,sizeof(ha2));
			memcpy(ha2.h,h+1,length-1);
			ha2.w1[0]=what1;
			ha2.w2[0]=what2;
			ha2.f=HASH_LEAF|1;
			a=ctx->han++*sizeof(struct hash_ha_t);
			if(lseek(ctx->f,a,SEEK_SET)!=a) {
				qerror(EESEEK,"in hash_hash_insert");
				return(-1);
			};
			if((w=safe_write(ctx->f,&ha2,sizeof(ha2)))!=sizeof(ha2)){
				qerror(EEWRITE,"in hash_hash_insert");
				return(-1);
			};
			return(0);
		}
	} else {
		// this must be a leaf, which cannot be split:
		// we have to add a collision entry.
		// !!hash is always 1 here
		int foo;
		if(((foo=ctx->ha[1].f&HASH_IDC))<HASH_MAX_I){
			(*(struct hash_hb_t *)&ctx->ha[1]).w1[foo]=what1;
			(*(struct hash_hb_t *)&ctx->ha[1]).w2[foo]=what2;
			ctx->ha[1].f++;
			if(hash_i_flush(ctx)<0);
			return foo;
		} else {
			// too many collisions
			return(-2);
		}
	}
};

int hash_init(struct hash_ctx * ctx, char * file)
{
	off_t r,w;
	
	ctx->hacd=0;
	ctx->ha0d=0;

	if((ctx->f=open(file,O_RDWR|O_CREAT,0666))<0){
		qerror2(EEOPEN,"in hash_init: ",file);
		return(-1);
	}
	
	if((r=safe_read(ctx->f,&ctx->ha[0],sizeof(ctx->ha[0])) )==0){
		bzero(&ctx->ha[0],sizeof(ctx->ha[0]));
		if((w=safe_write(ctx->f,&ctx->ha[0],sizeof(ctx->ha[0])))!=sizeof(ctx->ha[0])) {
		qerror2(EEWRITE,"in hash_init: ",file);
			close(ctx->f);
			return(-1);
		};
	} else if(r!=sizeof(ctx->ha[0])){
		// damaged index file
			qerror_m(EEREAD,"in hash_init: damaged hash file",file);
		close(ctx->f);
		return(-1);
	};
	ctx->han=(r=lseek(ctx->f,0,SEEK_END))/sizeof(ctx->ha[0]);
	if(r%sizeof(ctx->ha[0])) {
		qerror_m(EEWRONGOPT,"in hash_init: damaged hash file",file);
		return(-1);
	}
	ctx->hac=-1;
	return 0;
}

int hash_close(struct hash_ctx * ctx, char * file)
{
	if((ctx->f=open(file,O_RDWR|O_CREAT,0666))<0){
		qerror2(EECLOSE,"in hash_close: ", file);
		return(-1);
	}
	return(0);
}
